from typing import TypedDict
from uuid import UUID

from typing_extensions import NotRequired


class BrokerMessageHeader(TypedDict):
    dst: str
    src: str


class BrokerMessageStatus(TypedDict):
    code: int
    message: str


class BrokerMessage(TypedDict):
    request_type: str
    request_id: UUID
    header: BrokerMessageHeader
    body: dict


class UnprocessedBrokerMessage(BrokerMessage):
    status: NotRequired[BrokerMessageStatus]


class ProcessedBrokerMessage(BrokerMessage):
    status: BrokerMessageStatus



import logging
from abc import ABC, abstractmethod

from rabbit_broker.settings import RABBIT_URL

logger = logging.getLogger(__name__)


class AsyncAbstractMessageQueue(ABC):
    MessageQueue: str = ""

    def __init__(self):
        self.broker_url = RABBIT_URL
        self.connection = None
        self.client_properties = None
        logger.debug("%s.%s: Initialized", self.__class__.__name__, self.__init__.__name__)

    @abstractmethod
    async def __aenter__(self):
        pass

    @abstractmethod
    async def __aexit__(self, *args, **kwargs):
        pass

    @abstractmethod
    async def register_tasks(self, routing_key, worker):
        pass

    @abstractmethod
    async def consume(self):
        pass

    @abstractmethod
    async def post_message(self):
        pass



class Singleton(object):
    _instances = {}

    def __new__(cls, *args, **kwargs):
        if cls.__name__ not in cls._instances:
            cls._instances[cls.__name__] = super().__new__(cls, *args, **kwargs)
        return cls._instances[cls.__name__]



"""Данные pydantic модели - могут быть использованы как для валидации,
так и для генерации сообщения.

Валидация сообщения происходит так же, как и в обычной pydantic модели:

    ProcessedMessage(**message_dict)

Для генерации сообщения необходимо создать объект модели, не передавая аргументы
при инициализации, и вызвать метод generate. В аргументы метода generate можно
передавать любой из ключей структуры сообщения, в том числе code, message,
dst и src. Вложенная структура(header, status) формируется сама:

    >>> ProcessedMessage().generate(dst="destination", code=201, request_type="creation")
    >>> {"header": {"src": "", "dst": "destination"}, "request_type": "creation"...}
"""

from typing import Iterable, Optional, Union
from uuid import UUID, uuid4

from pydantic import BaseModel, Field


class MessageHeader(BaseModel):
    src: str
    dst: str


class MessageStatus(BaseModel):
    message: str
    code: int


class BaseMessage(BaseModel):
    request_type: str
    request_id: UUID
    body: Union[dict, Iterable]
    header: MessageHeader
    status: MessageStatus

    def __init__(self, **kwargs):
        """При вызове метода генерации сообщения, нужно создать экземляр модели,
        не инициализируя её и не передавая аргументы.
        Валидация происходит как в стандарной pydantic модели.
        """
        if kwargs:
            super().__init__(**kwargs)

    def generate(self, **fields) -> dict:
        """Генерирует сообщение."""
        flat_message = self.generate_flat_message(**fields)
        return self.get_structured_message(flat_message)

    def generate_flat_message(self, **fields) -> dict:
        """Заполняет плоскую структуру сообщения переданными значениями.
        Если значение не указано - берет его из DefaultValues.
        """
        flat_message = dict()
        for field_name, default_value in self.DefaultValues().dict().items():
            if value := fields.get(field_name):
                if isinstance(default_value, str):
                    value = str(value)
                flat_message[field_name] = value
            else:
                flat_message[field_name] = default_value
        return flat_message

    def get_structured_message(self, flat_message: dict) -> dict:
        """Создает вложенность в плоском сообщении."""
        flat_message["header"] = {
            "dst": flat_message.pop("dst"),
            "src": flat_message.pop("src"),
        }

        flat_message["status"] = {
            "message": flat_message.pop("message"),
            "code": flat_message.pop("code"),
        }
        for field_name in self.Config.to_exclude:
            flat_message.pop(field_name)

        return flat_message

    def dict(
        self,
        *,
        include=None,
        exclude=None,
        by_alias=False,
        exclude_unset=False,
        exclude_defaults=False,
        exclude_none=True,
    ):
        """Всегда исключать ключи со значением None."""
        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

    class Config:
        """Ключи, которые нужно исключить после генерации сообщения."""

        to_exclude = []

    class DefaultValues(BaseModel):
        """Значения по умолчанию при генерации сообщения."""

        request_id: UUID = Field(default_factory=uuid4)
        request_type: str = ""
        body: dict = dict()
        src: str = ""
        dst: str = ""
        message: str = "OK"
        code: int = 200


class ErrorMessage(BaseMessage):
    class DefaultValues(BaseMessage.DefaultValues):
        code: int = 400
        message: str = "Error"


class UnprocessedMessage(BaseMessage):
    status: Optional[MessageStatus] = None

    class Config:
        to_exclude = [
            "status",
        ]


class ProcessedMessage(BaseMessage):
    pass



from typing import List, Optional

from pydantic import BaseModel, Field


class PersistentQueueModel(BaseModel):
    queues_to_declare: Optional[List[str]] = Field(
        default_factory=list, description="Queue that will be created, if not existe"
    )



import logging
from abc import ABC, abstractmethod
from typing import Dict

from pydantic import ValidationError

from rabbit_broker.models import ErrorMessage, ProcessedMessage, UnprocessedMessage
from rabbit_broker.schemas import (
    BrokerMessageHeader,
    ProcessedBrokerMessage,
    UnprocessedBrokerMessage,
)
from rabbit_broker.utils.singleton import Singleton

logger = logging.getLogger(__name__)


class AbstractChain(ABC):
    chains: Dict[str, "BaseChain"] = {}

    def add(self, chain: "BaseChain") -> None:
        """Добавляет нового обработчика в цепочку."""
        self.chains[chain.request_type.lower()] = chain
        logger.debug(
            "%s.%s: %s added to chains.",
            self.__class__.__name__,
            self.add.__name__,
            chain.__name__,
        )

    @abstractmethod
    async def handle(self, data: UnprocessedBrokerMessage) -> ProcessedBrokerMessage: ...

    @abstractmethod
    def get_response_header(self, data: UnprocessedBrokerMessage) -> BrokerMessageHeader: ...  # pragma: no cover

    @abstractmethod
    async def get_response_body(self, data: UnprocessedBrokerMessage) -> ProcessedBrokerMessage: ...  # pragma: no cover

    def form_response(
        self,
        data: UnprocessedBrokerMessage,
        body: dict = None,
        code: int = 200,
        message: str = "",
    ) -> ProcessedBrokerMessage:
        body = body or {}
        data.update({"body": body})
        data.update({"status": {"message": str(message), "code": code}})
        logger.debug(
            "%s.%s: Formed response data=%s",
            self.__class__.__name__,
            self.form_response.__name__,
            data,
        )
        return data


class BaseChain(AbstractChain):
    """
    Базовый классов обработчиков.

    Args:
        AbstractChain: Интерфейс классов обработчиков.

    Attributes:
        request_type (str): Тип запроса, который обработчик способен обработать.
        include_in_schema (bool): True (значение по умолчанию) - выводить Chain в Swagger документацию;
                                False - исключить Chain из Swagger документации.
        deprecated (bool): False (значение по умолчанию) - Chain актуален;
                        True - отметить Chain, как устаревший.
        actual (str): Наименование актуального Chain в Swagger документации. Отображается
                    рядом с устаревшим Chain (где include_in_schema = True, deprecated = True).
                    Устанавливает deprecated = True автоматически, если deprecated не был указан как True.
    """

    request_type: str = ""
    include_in_schema: bool = True
    deprecated: bool = False
    actual: str = ""

    async def handle(self, data: UnprocessedBrokerMessage) -> ProcessedBrokerMessage:
        """
        Обрабатывает запрос, пропуская его через методы обработки
        заголовка и тела запроса.

        Args:
            data (dict): Словарь с запросом.

        Returns:
            Обработанный запрос: если типы запроса переданного сообщения
            и конкретного экземпляра обработчика совпадают.

            Метод handle() у родительского класса: если типы запроса переданного сообщения
            и конкретного экземпляра обработчика отличаются.
        """
        logger.info("%s.%s: data=%s", self.__class__.__name__, self.handle.__name__, data)
        try:
            UnprocessedMessage(**data)
        except ValidationError as error:
            logger.error("%s.%s: %s", self.__class__.__name__, self.handle.__name__, str(error))
            return ErrorMessage().generate(message=str(error))
        if self.request_type.lower() == data["request_type"].lower():
            response = ProcessedMessage().generate()
            try:
                response.update(await self.get_response_body(data))
                logger.debug(
                    "%s.%s: After body update response=%s",
                    self.handle.__name__,
                    self.__class__.__name__,
                    response,
                )
            except Exception as exc:
                return ErrorMessage().generate(message=str(exc))
            response.update(self.get_response_header(data))
            logger.debug(
                "%s.%s: After header update response=%s",
                self.__class__.__name__,
                self.handle.__name__,
                response,
            )
            # These field must stay the same.
            response["request_id"] = data["request_id"]
            response["request_type"] = data["request_type"]
            logger.info(
                "%s.%s: Before sending response=%s",
                self.__class__.__name__,
                self.handle.__name__,
                response,
            )
            try:
                ProcessedMessage(**response)
                return response
            except ValidationError as error:
                logger.error(
                    "%s.%s: ValidationError: %s",
                    self.__class__.__name__,
                    self.handle.__name__,
                    str(error),
                )
                return ErrorMessage().generate(message=str(error))
        else:
            logger.error(
                "%s.%s: Unknown request_type=%s",
                self.__class__.__name__,
                self.handle.__name__,
                data["request_type"],
            )
            return ErrorMessage().generate(message="Can't handle this request type")

    def get_response_header(self, data: UnprocessedBrokerMessage) -> BrokerMessageHeader:
        """Меняет местами получателя('dst') и отправителя('src') запроса."""
        updated_header = {"header": {"src": data["header"]["dst"], "dst": data["header"]["src"]}}
        logger.debug(
            "%s.%s: updated_header=%s",
            self.__class__.__name__,
            self.get_response_header.__name__,
            updated_header,
        )
        return updated_header


class ChainManager(BaseChain, Singleton):
    """Единая точка для распределения запросов по обработчикам."""

    chains = {}

    def __init__(self, parent_chain: BaseChain = BaseChain) -> None:
        """Собирает все обработчики в словарь."""
        if subclasses := parent_chain.__subclasses__():
            for subclass in subclasses:
                if subclass.request_type:
                    self.add(subclass)
                self.__init__(subclass)

    async def handle(self, data: UnprocessedBrokerMessage) -> ProcessedBrokerMessage:
        """Направляет запрос на нужный обработчик."""
        try:
            UnprocessedMessage(**data)
            chain = self.chains[data["request_type"].lower()]
            return await chain().handle(data)
        except ValidationError as error:
            msg = f"Incoming message validation error: {error}"
        except KeyError as error:
            msg = f"Can't handle this request type: {error}"
        logger.error("%s.%s: %s", self.__class__.__name__, self.handle.__name__, msg)
        return ErrorMessage().generate(message=msg)

    async def get_response_body(self, data):
        pass



import asyncio
import logging

import aio_pika
from aio_pika.patterns import RPC
from pydantic import ValidationError

from rabbit_broker.models import ErrorMessage, ProcessedMessage, UnprocessedMessage
from rabbit_broker.queues.base import AsyncAbstractMessageQueue
from rabbit_broker.schemas import ProcessedBrokerMessage, UnprocessedBrokerMessage

logger = logging.getLogger(__name__)


class AsyncRabbitMessageQueue(AsyncAbstractMessageQueue):
    async def consume(self) -> None:
        logger.info(
            "%s.%s: RPC consumer started",
            self.__class__.__name__,
            self.consume.__name__,
        )
        await asyncio.Future()

    async def post_message(self, data: UnprocessedBrokerMessage, worker: str) -> ProcessedBrokerMessage:
        try:
            UnprocessedMessage(**data)
        except ValidationError as error:
            logger.error(
                "%s.%s: UnprocessedMessage validation failed!: %s",
                self.__class__.__name__,
                self.post_message.__name__,
                str(error),
            )
            return ErrorMessage().generate(message=str(error))
        response = await self.rpc.call(worker, kwargs=dict(data=data))
        try:
            ProcessedMessage(**response)
        except ValidationError as error:
            return ErrorMessage().generate(
                request_id=data["request_id"],
                request_type=data["request_type"],
                dst=data["src"],
                src=data["dst"],
                message=str(error),
            )
        return response

    async def register_tasks(self, routing_key: str, worker: callable):
        """Вызывать перед стартом консьюмера."""
        await self.rpc.register(routing_key, worker, auto_delete=True)

    async def __aenter__(self):
        """
        Метод входа в контекст подключения
        """
        if self.connection is None or self.connection.is_closed:
            logger.info(
                "%s.%s: Created connection",
                self.__class__.__name__,
                self.__aenter__.__name__,
            )
            self.connection = await aio_pika.connect_robust(
                self.broker_url,
            )
            self.channel = await self.connection.channel()
            self.rpc = await RPC.create(self.channel)
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.connection.close()
        await self.channel.close()



import logging
from abc import abstractmethod

from pydantic import ValidationError

from rabbit_broker.async_chains.base import BaseChain as AsyncBaseChain
from rabbit_broker.async_chains.base import ChainManager as AsyncChainManager
from rabbit_broker.models import ErrorMessage, ProcessedMessage, UnprocessedMessage
from rabbit_broker.schemas import ProcessedBrokerMessage, UnprocessedBrokerMessage
from rabbit_broker.utils.singleton import Singleton

logger = logging.getLogger(__name__)


class BaseChain(AsyncBaseChain):
    """Синхронная версия базового класса обработчика."""

    def handle(self, data: UnprocessedBrokerMessage) -> ProcessedBrokerMessage:
        """
        Обрабатывает запрос, пропуская его через методы обработки
        заголовка и тела запроса.

        Args:
            data (dict): Словарь с запросом.

        Returns:
            Обработанный запрос: если типы запроса переданного сообщения
            и конкретного экземпляра обработчика совпадают.

            Метод handle() у родительского класса: если типы запроса переданного сообщения
            и конкретного экземпляра обработчика отличаются.
        """
        logger.info("%s.%s: data=%s", self.__class__.__name__, self.handle.__name__, data)
        try:
            UnprocessedMessage(**data)
        except ValidationError as error:
            logger.error("%s.%s: %s", self.__class__.__name__, self.handle.__name__, str(error))
            return ErrorMessage().generate(message=str(error))
        if self.request_type.lower() == data["request_type"].lower():
            response = ProcessedMessage().generate()
            try:
                response.update(self.get_response_body(data))
                logger.debug(
                    "%s.%s: After body update response=%s",
                    self.__class__.__name__,
                    self.handle.__name__,
                    response,
                )
            except Exception as exc:
                return ErrorMessage().generate(message=str(exc))
            response.update(self.get_response_header(data))
            logger.debug(
                "%s.%s: After header update response=%s",
                self.__class__.__name__,
                self.handle.__name__,
                response,
            )
            # These field must stay the same.
            response["request_id"] = data["request_id"]
            response["request_type"] = data["request_type"]
            logger.info(
                "%s.%s: Before sending response=%s",
                self.__class__.__name__,
                self.handle.__name__,
                response,
            )
            try:
                ProcessedMessage(**response)
                return response
            except ValidationError as error:
                logger.error(
                    "%s.%s: ValidationError: %s",
                    self.__class__.__name__,
                    self.handle.__name__,
                    str(error),
                )
                return ErrorMessage().generate(message=str(error))
        else:
            logger.error(
                "%s.%s: Unknown request_type=%s",
                self.__class__.__name__,
                self.handle.__name__,
                data["request_type"],
            )
            return ErrorMessage().generate(message="Can't handle this request type")

    @abstractmethod
    def get_response_body(self, data: UnprocessedBrokerMessage) -> ProcessedBrokerMessage: ...


class ChainManager(AsyncChainManager, Singleton):
    """Синхронная версия менеджера распределения запросов."""

    def handle(self, data: UnprocessedBrokerMessage) -> ProcessedBrokerMessage:
        """Направляет запрос на нужный обработчик."""
        try:
            UnprocessedMessage(**data)
            chain = self.chains[data["request_type"].lower()]
            return chain().handle(data)
        except ValidationError as error:
            msg = f"Incoming message validation error: {error}"
        except KeyError as error:
            msg = f"Can't handle this request type: {error}"
        logger.error("%s.%s: %s", self.__class__.__name__, self.handle.__name__, msg)
        return ErrorMessage().generate(message=msg)

    def get_response_body(self, data):
        pass



import json
import logging
from abc import ABC, abstractmethod
from typing import Union

import aio_pika

from rabbit_broker.queues.models import PersistentQueueModel
from rabbit_broker.settings import RABBIT_URL

logger = logging.getLogger(__name__)


class AsyncAbstractPersistentQueue(ABC):
    def __init__(self, properties: PersistentQueueModel = None):
        self.broker_url = RABBIT_URL
        self.connection = None
        self.properties = properties or PersistentQueueModel()
        logger.debug("%s.%s: Initialized", self.__class__.__name__, self.__init__.__name__)

    @abstractmethod
    async def __aenter__(self):
        pass

    @abstractmethod
    async def __aexit__(self, *args, **kwargs):
        pass

    @abstractmethod
    async def post_message(self):
        pass


class AsyncPersistentQueue(AsyncAbstractPersistentQueue):
    async def __aenter__(self):
        if self.connection is None or self.connection.is_closed:
            self.connection = await aio_pika.connect_robust(self.broker_url)
            logger.info("%s.%s: Created robust connection", self.__class__.__name__, self.__aenter__.__name__)
            self.channel = await self.connection.channel()

            for queue_name in self.properties.queues_to_declare:
                await self.channel.declare_queue(queue_name, durable=True)
                await self.channel.declare_queue(queue_name, durable=True)

        return self

    async def __aexit__(self, *args, **kwargs):
        await self.connection.close()
        await self.channel.close()

    async def post_message(self, queue_name: str, data: dict) -> None:
        json_body = json.dumps(data).encode("utf-8")
        message = aio_pika.Message(body=json_body, delivery_mode=aio_pika.DeliveryMode.PERSISTENT)
        await self.channel.default_exchange.publish(message, routing_key=queue_name)

    async def peek_message(self, queue_name) -> Union[str, None]:
        queue = await self.channel.get_queue(queue_name)
        message = await queue.get(no_ack=False)

        decoded_message = None

        if message:
            decoded_message = message.body.decode("utf-8")
            await message.nack(requeue=True)

        return decoded_message



import asyncio
import logging

import aio_pika
from aio_pika.patterns import RPC
from pydantic import ValidationError

from rabbit_broker.models import ErrorMessage, ProcessedMessage, UnprocessedMessage
from rabbit_broker.schemas import ProcessedBrokerMessage, UnprocessedBrokerMessage
from rabbit_broker.settings import RABBIT_URL, SERVICE_NAME

logger = logging.getLogger(__name__)


class BaseService:
    """Отправка сообщений в сервисы."""

    broker_url = RABBIT_URL
    service_name = SERVICE_NAME

    def __init__(self):
        """Создает необходимые атрибуты для подключения к брокеру сообщений."""
        if not self.dst_service_name:
            raise AttributeError(f"Attribute `dst_service_name` has not been set for class {self.__class__.__name__}")

    async def send_message(self, request_type: str, body: dict) -> ProcessedBrokerMessage:
        """Генерирует уникальный id запроса и вызывает отправку сформированного
        сообщения.
        """
        message = UnprocessedMessage().generate(
            request_type=request_type,
            src=self.service_name,
            dst=self.dst_service_name,
            body=body,
        )
        return await self.send_rpc_request(message)

    async def send_rpc_request(self, message: UnprocessedBrokerMessage) -> ProcessedBrokerMessage:
        """Валидирует сообщение, создает соединение с брокером и отправляет
        сообщение в очередь.
        В случае ошибки формирует сообщение с данными об ошибке и HTTP кодом 400.
        """
        try:
            UnprocessedMessage(**message)
        except ValidationError as error:
            logger.error(
                "%s.%s: UnprocessedMessage validation failed!: %s",
                self.__class__.__name__,
                self.send_rpc_request.__name__,
                str(error),
            )
            return ErrorMessage().generate(message=str(error))
        try:
            connection = await aio_pika.connect_robust(self.broker_url)
            async with connection, connection.channel() as channel:
                rpc = await RPC.create(channel)
                response = await rpc.call(self.dst_service_name, kwargs=dict(data=message))
                ProcessedMessage(**response)
                return response
        except ValidationError as error:
            # Временный фикс, пока все сервисы не перейдут на новую версию пакета.
            logger.error(
                "%s.%s: UnprocessedMessage validation failed!: %s",
                self.__class__.__name__,
                self.send_rpc_request.__name__,
                str(error),
            )
            return response
        except (
            asyncio.TimeoutError,
            asyncio.CancelledError,
            RuntimeError,
        ) as err:
            return ErrorMessage().generate(
                request_id=message["request_id"],
                request_type=message["request_type"],
                src=self.dst_service_name,
                dst=self.service_name,
                message=str(err),
            )
